import yaml
import json
import toml as toml


def dictify(absolute_file_path: str, toml_var: str = "META_TOML"):
    """Helper script to extract data from .json, .py, .yaml, .toml files and loads into a python dict

    :param absolute_file_path: absolute path to file to parse data from
    :param toml_var: name of variable you store a toml string into (optional)

    :return: dictionary extracted by dictify
    """
    toml_contents = ""
    derived_metadata = dict()
    if absolute_file_path.endswith('.yaml') or absolute_file_path.endswith('.yml'):
        with open(absolute_file_path, 'r') as yml_file:
            yaml_contents = yaml.safe_load(yml_file)
            toml_contents = toml.dumps(yaml_contents)
    elif absolute_file_path.endswith('.json'):
        with open(absolute_file_path, 'r') as json_file:
            json_contents = json.loads(json_file.read())
            toml_contents = toml.dumps(json_contents)
    elif absolute_file_path.endswith('.toml'):
        with open(absolute_file_path, 'r') as toml_file:
            toml_contents = toml.dumps(toml.load(toml_file))
    elif absolute_file_path.endswith('.py'):
        toml_contents = load_toml_from_meta_toml_str(absolute_file_path=absolute_file_path, toml_var=toml_var)
    try:
        derived_metadata = toml.loads(toml_contents)
    except toml.TomlDecodeError:
        print("Error decoding TOML from file")
    return derived_metadata


def load_toml_from_meta_toml_str(absolute_file_path: str, toml_var: str = "META_TOML"):
    """Simple approach to parse out metadata toml declared in a string

    :param absolute_file_path: absolute path to file to parse data from
    :param toml_var: name of variable you store a toml string into (defaulted)

    :rtype: string
    """
    try:
        toml_contents: str = ""
        with open(absolute_file_path, 'r') as file:
            line = file.readline()
            while line:
                if toml_var in line:
                    line = file.readline()
                    while '\"\"\"' not in line:
                        toml_contents += line
                        line = file.readline()
                    break
                else:
                    line = file.readline()
    except Exception as exc:
        print(f"Error while trying to read file's metadata, exception: \n {exc}")
    finally:
        return toml_contents



from subprocess import check_call, check_output
from sys import executable
from os import path
from autometa.util import dictify


class Autometa:
    """ An object to hold necessary attributes for parsing a variety of metadata
    """

    __attrs__ = [
        "source_file_path",
        "exclusions_file_path",
        "dependencies",
        "exclusions",
        "metadata"
        ]

    def __init__(self, absolute_file_path: str, dependencies=None, exclusions_files_path: str = None,
                 exclusions=None):
        """

        :param absolute_file_path: required to instantiate object, path to file to parse from
            #TODO: make this recursively append files from directory instead of 1 file?
        :param dependencies: can be added manually or will be populated later using class functions
        :param exclusions_files_path: optional filepath to excluded_from_uninstall packages
        :param exclusions: can be added manually in instantiation as a list of strings, or populated by class
        """

        if dependencies is None:
            dependencies = []
        if exclusions is None:
            exclusions = []
        try:
            if path.exists(absolute_file_path):
                self.source_file_path = absolute_file_path
                self.dependencies = dependencies
                self.exclusions_files_path = exclusions_files_path
                self.exclusions = exclusions
                self.metadata = None
            else:
                raise Exception(f"Source file path does not exist.")
        except Exception as exc:
            raise Exception(f"Exception raised: {exc}")

    def get_source_file_path(self):
        sfp = self.source_file_path
        return sfp

    def get_metadata(self):
        md = self.metadata
        return md

    def get_dependencies(self):
        gd = self.dependencies
        return gd

    def get_exclusions(self):
        ge = self.exclusions
        return ge

    def set_metadata(self, metadata):
        self.metadata = metadata

    def set_dependencies(self, dependencies: list):
        self.dependencies = dependencies

    def update_metadata(self, toml_var: str = "META_TOML"):
        """ set metadata toml from file into a variable for use by autometa object

        :param toml_var: variable name that is storing metadata toml string
        """
        derived_toml = dictify(absolute_file_path=self.get_source_file_path(), toml_var=toml_var)
        self.set_metadata(derived_toml)

    def parse_dependencies(self, toml_table_key: str = "project", toml_table_value: str = "dependencies"):
        """

        :param toml_table_value: variable name in toml that you want to pull a list from
        :param toml_table_key: subheading / table name in TOML holding the dependencies variable

        :return: list of strings, pip-package install-list if present
        """
        dependencies = []
        if not self.metadata:
            self.update_metadata()
            print(f"metadata wasn't populated, doing now... {self.metadata}")
        derived_metadata = self.get_metadata()
        if derived_metadata:
            try:
                if toml_table_key in derived_metadata and toml_table_value in derived_metadata[toml_table_key]:
                    dependencies = derived_metadata[toml_table_key][toml_table_value]
            except Exception as exc:
                raise Exception(f"Failed to set dependencies, Exception: {exc}")
            if dependencies:
                self.set_dependencies(dependencies)
                print(f"Dependencies successfully parsed: {self.get_dependencies()}")
            else:
                print("Dependencies list is empty")

    def pip_install_dependencies(self, dependencies: list):
        """ attempts to pip install packages in dependencies

        :param dependencies: list of pypi package names to install
        """
        for dependency in self.get_dependencies():
            dependencies.append(dependency)
        for dependency in dependencies:
            check_call([executable, '-m', 'pip', 'install', dependency])
        reqs = check_output([executable, '-m', 'pip', 'freeze'])

        installed_packages = [r.decode().split("==")[0] for r in reqs.split()]
        installed_packages = [r.lower() for r in installed_packages]
        for dependency in dependencies:
            if dependency.lower() not in installed_packages:
                raise Exception(f"failed to pip install {dependency}")

    def pip_uninstall_dependencies(self, dependencies: list, exclusions: list):
        """attempts to pip uninstall packages listed

        :param exclusions: list of pypi packages to never uninstall
        :param dependencies: list of pypi package names to install
        """
        for dependency in self.get_dependencies():
            dependencies.append(dependency)
        for dependency in self.get_exclusions():
            exclusions.append(dependency)
        uninstall_list = []

        for dependency in dependencies:
            dependency = dependency.lower()
            if dependency not in exclusions:
                uninstall_list.append(dependency)

        for dependency in uninstall_list:
            check_call([executable, '-m', 'pip', 'uninstall', dependency, '-y'])
        reqs = check_output([executable, '-m', 'pip', 'freeze'])

        installed_packages = [r.decode().split("==")[0] for r in reqs.split()]
        installed_packages = [r.lower() for r in installed_packages]
        for dependency in uninstall_list:
            if dependency in installed_packages and dependency in uninstall_list:
                raise Exception(f"failed to pip uninstall {dependency}")
